/* this program is designed to convert my neoletters font into a ttf */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <stdarg.h>
#include "u8.h"
#include "unidat.h"
#include "bdf.h"
#include "oio.h"
#include "ttf.h"
#include "etc.h"

#define ei else if
int xdgval(int c){
	if(c >= '0' && c <= '9')return c - '0';
	if(c >= 'A' && c <= 'F')return c - 'A' + 10;
	if(c >= 'a' && c <= 'f')return c - 'a' + 10;
	return 0;
}

unsigned ulog2(unsigned x){
	asm("bsrl %0,%0":"+r"(x));
	return x;
}

void makepathmap(struct bdfglyph*g);
//temporary storage;
char line[300];
char pathmap[17*19];
struct point {signed char x,y,start;} pathlist[17*19*2];
int pathend;

int bdfpthchk(struct bdfglyph *g,int x,int y);
void addpath(int x,int y);

struct table_entry {
	char name[4];
	unsigned info_loc;
	unsigned loc;
	unsigned len;
	unsigned state;//0 not written, 1 writing, 2 finished
};

struct ttf_table {
	unsigned size;
	struct table_entry * table_entries;
};

int table_entry_compar(const void *av, const void *bv){
	return strncmp(av,bv,4);
}

void sort_ttf_table(struct ttf_table*ttf) {
	qsort(ttf->table_entries, ttf->size, sizeof(struct table_entry), table_entry_compar);
}

struct table_entry *find_table_entry(struct ttf_table *ttf,const char*s,int state){
	struct table_entry *r=bsearch(s, ttf->table_entries, ttf->size, sizeof(struct table_entry), table_entry_compar);
	if(r==0){printf("ERROR table %s not found\n",s);exit(1);}
	r->state = state;
	return r;
}

void add_table_names(struct ttf_table *ttf, const char*s) {
	int l=strlen(s);
	while(l >= 4){
		int z=ttf->size;
		memcpy(ttf->table_entries[z].name,s,4);
		ttf->table_entries[z].info_loc=0;
		ttf->table_entries[z].loc=0;
		ttf->table_entries[z].len=0;
		ttf->table_entries[z].state=0;
		ttf->size++;
		s += 5; l -= 5;
	}
	sort_ttf_table(ttf);
}

void write_blank_tables(FILE *f,struct ttf_table *ttf){
	unsigned i;
	for (i=0; i < ttf->size; i++){
		struct table_entry *ent=(ttf->table_entries) + i;
		writestr(f,ent->name);
		ent->info_loc=ftell(f);
		write32(f,0);//cksum
		write32(f,0);//location
		write32(f,0);//length
	}
}

void table_status_disp(struct ttf_table *ttf){
	printf(GOLN(3));
	int i;
	for (i=0; i < ttf->size; i++){
		struct table_entry *ent=(ttf->table_entries) + i;
		if(ent->state == 0)printf(TYEL"[%.4s]"CLRCOL,ent->name);
		if(ent->state == 1)printf(TGRN"[%.4s]"CLRCOL,ent->name);
		if(ent->state == 2)printf(TBCYN"[%.4s]"CLRCOL,ent->name);
	}
}

void finalizetable(FILE *f,struct table_entry *ent){
	int padc=ent->len % 4;
	if(padc==1)fwrite("\0\0\0",1,3,f);
	if(padc==2)fwrite("\0\0",1,2,f);
	if(padc==3)fwrite("\0",1,1,f);
	unsigned lenwpad=ent->len/4 + (padc?1:0);
	uint32_t sum=0;
	goloc(f,ent->loc);
	unsigned i;
	for(i=0;i<lenwpad;i++)sum += read32(f);
	goloc(f,ent->info_loc);
	write32(f,sum);
	write32(f,ent->loc);
	write32(f,ent->len);
	ffastfw(f);
	if(ftell(f) % 4 != 0){printf("ERROR table misaligned\n");exit(1);}
	ent->state = 2;
}

void bdf2ttf_status(struct bdfglyph *g,int cur,int tot) {
	printf(GOLN(8)CLRLIN"status: U%05x\n\r"CLRLIN"name: %s",get_glyph_encoding(g),get_glyph_name(g));
	printf(GOLN(3)"%3d%%"TCYN BYEL,cur*100/tot);
	oio_showbar(tot,cur,30);
	printf(CLRCOL);
	outputasbitmap(g,4,1);
}

int main (int argc, char**argv) {
	printf(CLRSCR TOPLEF BRED "Oren's 𝔹𝔻𝔽 2 𝑻𝑻𝑭" CLRCOL "\n");

	const char * bdfname="in.bdf";
	const char * ttfname="out.ttf";
	if (argc > 1)bdfname=argv[1];
	if (argc > 2)ttfname=argv[2];
	checkext(bdfname,"bdf");
	checkext(ttfname,"ttf");

	FILE *log=fopen("bdf2ttf.log","w+");
	FILE *out=fopen(ttfname,"w+");
        if(!out){
		printf(TRED TOPLEF "Error, could not open output file" CLRCOL "\r\n");
		abort();
	}
	printf(GOLN(2)"Loading BDF %s",bdfname);

	struct bdffont *my_font = load_font_alloc(bdfname,bdf2ttf_status);

	if (!my_font){
		printf(TRED TOPLEF "Error, could not open input file" CLRCOL "\r\n");
		abort();
	}

	struct table_entry ttf_table_data[20];
	struct ttf_table ttf_tbl={0,ttf_table_data};
	add_table_names(&ttf_tbl, "OS/2 cmap cvt  glyf head hhea hmtx loca maxp name post");

	const char *font_name = get_font_name(my_font);
	const char *gsub_file = NULL;

	int argi = 3;
	while (argc > argi + 1){
		if(!strcmp(argv[argi],"-GSUB")){
			gsub_file = argv[argi+1];
			add_table_names(&ttf_tbl,"GSUB");
		}
		if(!strcmp(argv[argi],"-name")){
			font_name = argv[argi+1];
		}
		argi += 2;
	}

	load_eawdat();

	printf(GOLN(2)CLRBEL "Writing TTF Tables: %s",get_font_name(my_font));
	table_status_disp(&ttf_tbl);
	/* write ttf file header */
	write32(out,0x00010000);
	write16(out,ttf_tbl.size); //number of ttf tables
	int entryselector = ulog2(ttf_tbl.size);
	write16(out,16<<entryselector);
	write16(out,entryselector);
	write16(out,ttf_tbl.size*16-(16<<entryselector));
	write_blank_tables(out,&ttf_tbl);
	const int *bbx = get_font_bbx(my_font);

	unsigned UPM= bbx[1];
        int xmin= bbx[2];
        int xmax=xmin+bbx[0];
        int ymin=bbx[3];
        int ymax=ymin+bbx[1];

	struct table_entry *ent=find_table_entry(&ttf_tbl,"cvt ",1);
	table_status_disp(&ttf_tbl);
	ent->loc=ftell(out);
	write32(out,0x00000000);//AAAAAAAAA
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	fputs("cvt table finished\n",log);
	fflush(log);

	/* write the head table */
	ent=find_table_entry(&ttf_tbl,"head",1);
	table_status_disp(&ttf_tbl);
	ent->loc=ftell(out);
	write_head_table(out, UPM, xmin, ymin, xmax, ymax);
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	ent->state=1;
	table_status_disp(&ttf_tbl);

	fputs("head table finished\n",log);
	fflush(log);

	/*write the hhea table*/
	ent=find_table_entry(&ttf_tbl,"hhea",1);
	table_status_disp(&ttf_tbl);
	ent->loc=ftell(out);
        write_hhea_table(out, xmax, ymin, ymax, font_nchars(my_font));
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	fputs("hhea table finished\n",log);
	fflush(log);

	/*write the OS/2 table*/
	ent=find_table_entry(&ttf_tbl,"OS/2",1);
	table_status_disp(&ttf_tbl);
	ent->loc=ftell(out);
	char panose[10] = {2 /*Latin text*/,11/*normal sans serif*/,6/*medium weight*/,
		9/*monospaced*/,2 /*no contrast*/,2 /*no variation*/, 0,0,0,0};
	write_os_2_table(out, UPM, ymin, ymax,  8 /*sans serif*/, 3 /*geometric*/, panose,
		"OREN");
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	/*write the post table*/
	ent=find_table_entry(&ttf_tbl,"post",1);
	table_status_disp(&ttf_tbl);
	ent->loc=ftell(out);
	write_post_table(out);
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	/*write the name table*/
	ent=find_table_entry(&ttf_tbl,"name",1);
	table_status_disp(&ttf_tbl);
	ent->loc=ftell(out);
	write_name_table(out, font_name, "Regular", "Version 1.0");
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	/*loca table will be written during main loop*/
	ent=find_table_entry(&ttf_tbl,"loca",1);
	table_status_disp(&ttf_tbl);
	ent->loc=ftell(out);
	write32(out,0);
	/* open ttf table files */

	FILE *glyf=fopen("glyf.temp","w+");
	FILE *hmtx=fopen("hmtx.temp","w+");
	FILE *cmap12=fopen("cmap12.temp","w+");// format 12 table
	int startccode=-1;
	int endccode=-1;
	int startglyfid=0;
	int maxpoints=0;
	int maxcontour=0;
	unsigned glyfid=0;
	unsigned ngroups=0;
	unsigned n2groups=0;//number of groups below FFFF
	int ccode=-1;
        int bl=12;

	unsigned short *ccode2glyfid = malloc(sizeof(short)*0x20000);
	memset(ccode2glyfid,0xff,0x20000*sizeof(short));

	while(glyfid < font_nchars(my_font)){
		ccode++;
		while(ccode < 0x20000 && !get_glyph(my_font,ccode))ccode++;
		if(ccode >= 0x20000)break;
		struct bdfglyph *g=get_glyph(my_font,ccode);
		printf(GOLN(4)"U%05x, %d "TCYN BYEL,ccode,glyfid);
		oio_showbar(font_nchars(my_font),glyfid,30);
		printf(CLRCOL);
		int j,i;
		/* create path map */
		int xmin=20,xmax=0,ymin=16,ymax=0;
		for(i=0;i<=18;i++){
			for(j=0;j<=16;j++){
				pathmap[i+j*19]=bdfpthchk(g,i,j);
				if(pathmap[i+j*19] && pathmap[i+j*19] != 0xF){
					if(i<xmin)xmin=i;
					if(i>xmax)xmax=i;
					if(j<ymin)ymin=j;
					if(j>ymax)ymax=j;
				}
			}
		}
		/* use map to find all paths */
		pathend=0;
		int n=0;
		for(i=0;i<=18;i++){
			for(j=0;j<=16;j++){
				if(pathmap[i+j*19] && pathmap[i+j*19] != 0xF){
					addpath(i,j);
					n++;
				}
			}
		}
		if(n==0)goto outputloca;
		/* output paths into glyf */
		write16(glyf,n);//number of contours
		write16(glyf,xmin);//xmin
		write16(glyf,bl-ymax);//ymin
		write16(glyf,xmax);//xmax
		write16(glyf,bl-ymin);//ymax
		printf(GOLN(5)"paths: %d %d %d %d %d\n",n,xmin,bl-ymax,xmax,bl-ymin);
		for(i=1;i<pathend;i++){
			if(pathlist[i].start==1)write16(glyf,i-1);
		}
		write16(glyf,pathend-1);
		write16(glyf,0);// zero bytes of code
		for(i=0;i<pathend;i++){
			struct point p=pathlist[i];
			int flags=7;
			if(p.x==0)flags ^= 022;
			ei(p.x > 0)flags |= 020;
			if(p.y==0)flags ^= 044;
			ei(p.y > 0)flags |= 040;
			write8(glyf,flags);
		}
		for(i=0;i<pathend;i++){
			if(pathlist[i].x!=0)write8(glyf,abs(pathlist[i].x));
		}
		for(i=0;i<pathend;i++){
			if(pathlist[i].y!=0)write8(glyf,abs(pathlist[i].y));
		}
		if(pathend > maxpoints)maxpoints=pathend;
		if(n > maxcontour)maxcontour=n;
		while(ftell(glyf) & 3)write8(glyf,0);//align on 4 byte boundary
		outputloca:
		write32(out,ftell(glyf));
		//check if we need to output a character range record
		if(startccode==-1 && ccode != 0){startccode=endccode=ccode;startglyfid=glyfid;}
		if(endccode < ccode-1){
			if(endccode < 0x10000 && ccode > 0x10000)n2groups=ngroups +1;
			write32(cmap12,startccode);
			write32(cmap12,endccode);
			write32(cmap12,startglyfid);
			startccode=endccode=ccode;
			startglyfid=glyfid;
			ngroups++;
		}else endccode=ccode;
		//output hmtx
		if(ccode==0){
			write16(hmtx,9);
		} else {
			int w=get_eaw(ccode);
			if (w == 0)w = 1;
			write16(hmtx,9*w);
		}
		write16(hmtx,xmin==20?0:xmin);
		printf(GOLN(4)"U%05x, %d",ccode,glyfid);
		ccode2glyfid[ccode] = glyfid;
		glyfid++;
	}

	fputs("end of main loop\n",log);
	fflush(log);

	//final group
	write32(cmap12,startccode);
	write32(cmap12,endccode);
	write32(cmap12,startglyfid);
	ngroups++;

	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	ent=find_table_entry(&ttf_tbl,"hmtx",1);
	table_status_disp(&ttf_tbl);
	ent->loc=ftell(out);
	writefile(out,hmtx);
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	fputs("hmtx table finished\n",log);
	fflush(log);

	ent=find_table_entry(&ttf_tbl,"cmap",1);
	table_status_disp(&ttf_tbl);
	ent->loc=ftell(out);
	write_cmap_table(out, cmap12, n2groups, ngroups);
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	if(gsub_file != NULL){
		fputs("gsub table starting\n",log);
		fflush(log);

		ent=find_table_entry(&ttf_tbl,"GSUB",1);
		table_status_disp(&ttf_tbl);
		ent->loc=ftell(out);
		fputs("gsub table mid\n",log);
		fflush(log);
		write_GSUB_table(out, gsub_file, ccode2glyfid);
		ent->len=ftell(out) - ent->loc;
		finalizetable(out,ent);
		table_status_disp(&ttf_tbl);

		fputs("gsub table finished\n",log);
		fflush(log);

	}

	ent=find_table_entry(&ttf_tbl,"glyf",1);
	ent->loc=ftell(out);
	writefile(out,glyf);
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	ent=find_table_entry(&ttf_tbl,"maxp",1);
	ent->loc=ftell(out);
	write_maxp_table(out,glyfid,maxpoints,maxcontour);
	ent->len=ftell(out) - ent->loc;
	finalizetable(out,ent);
	table_status_disp(&ttf_tbl);

	printf(GOYX(7,1)"...computing final checksum...");
	uint32_t sum=0;
	unsigned len=ftell(out);
	goloc(out,0);
	int i;
	for(i=0;i<len;i++)sum += read32(out);
	goloc(out,find_table_entry(&ttf_tbl,"head",2)->loc + 8);
	write32(out,0xB1B0AFBA-sum);
	printf(GOYX(3,25)TBCYN"[head]"CLRCOL GOLN(8)"Finished!\n");
	free_bdffont(my_font);

	return 0;
}

struct point p0;
void addpath(int x,int y){
	int dir;
	int i=pathend;
	char map=pathmap[x+y*19];
	printf(GOLN(6)"\e[0Jstart at %d,%d:",x,y);
	if(i==0){
		pathlist[i]=(struct point){x,12-y,1};
		p0=(struct point){x,12-y,1};
	}else{
		pathlist[i]=(struct point){x-p0.x,12-y-p0.y,1};
		p0=(struct point){x,12-y,1};
	}
	if(map==8){dir='R';pathmap[x+y*19]=0;}
	if(map==7){dir='D';pathmap[x+y*19]=0;}
	i++;
	pathlist[i]=(struct point){0,0,0};
	nextpoint:;
	putchar(dir);
	fflush(stdout);
	int olddir=dir;
	if(dir=='R'){
		x++;
		pathlist[i].x++;
		char map=pathmap[x+y*19];
		if(map==4){dir='D';pathmap[x+y*19]=0;}
		ei(map==6){dir='D';pathmap[x+y*19]=2;}
		ei(map==12){dir='R';pathmap[x+y*19]=0;}
		ei(map==14){dir='U';pathmap[x+y*19]=0;}
		else goto endof;
	}ei(dir=='D'){
		y++;
		pathlist[i].y--;
		char map=pathmap[x+y*19];
		if(map==1){dir='L';pathmap[x+y*19]=0;}
		ei(map==5){dir='D';pathmap[x+y*19]=0;}
		ei(map==9){dir='L';pathmap[x+y*19]=8;}
		ei(map==13){dir='R';pathmap[x+y*19]=0;}
		else goto endof;
	}ei(dir=='L'){
		x--;
		pathlist[i].x--;
		char map=pathmap[x+y*19];
		if(map==2){dir='U';pathmap[x+y*19]=0;}
		ei(map==3){dir='L';pathmap[x+y*19]=0;}
		ei(map==6){dir='U';pathmap[x+y*19]=4;}
		ei(map==7){dir='D';pathmap[x+y*19]=0;}
		else goto endof;
	}ei(dir=='U'){
		y--;
		pathlist[i].y++;
		char map=pathmap[x+y*19];
		if(map==8){dir='R';pathmap[x+y*19]=0;}
		ei(map==9){dir='R';pathmap[x+y*19]=1;}
		ei(map==10){dir='U';pathmap[x+y*19]=0;}
		ei(map==11){dir='L';pathmap[x+y*19]=0;}
		else goto endof;
	}
	if(dir != olddir){
		p0.x += pathlist[i].x;
		p0.y += pathlist[i].y;
		i++;
		pathlist[i]=(struct point){0,0,0};
	}
	goto nextpoint;
	endof:;
	printf(" end at %d,%d\n",p0.x,p0.y);
	pathend=i; // the last segment is not included.
}


int bdfpixchk(struct bdfglyph *g,int x,int y){
	return get_glyph_pixel(g,y,x);
}

int bdfpthchk(struct bdfglyph *g,int x,int y){
	return	bdfpixchk(g,x-1,y-1) +
		2*bdfpixchk(g,x,y-1) +
		4*bdfpixchk(g,x-1,y) +
		8*bdfpixchk(g,x,y);
}
