/* OIO 0.1 */

#include <stdio.h>
#include <string.h>
#include <termios.h>
#include "oio.h"
#include "u8.h"

struct termios oio_oldterm;
struct termios oio_rawterm;

#define OIO_CHARBUFFER_SIZE 10
char oio_charbuffer[OIO_CHARBUFFER_SIZE];
int oio_charbuffer_index = 0;

unsigned oio_error_flag;

int oio_getchar() {
	if (oio_charbuffer_index > 0) {
		oio_charbuffer_index -= 1;
		return oio_charbuffer[oio_charbuffer_index];
	}
	return getchar();
}

void oio_ungetc(int c) {
	if (c == EOF) return; // eof is eof
	if (oio_charbuffer_index < OIO_CHARBUFFER_SIZE) {
		oio_charbuffer[oio_charbuffer_index] = c;
		oio_charbuffer_index += 1;
	} else {
		oio_error_flag |= OIO_ERROR_BUFFER_OVERFLOW;
	}
}


void oio_setup () {
	tcgetattr(0,&oio_oldterm);
	cfmakeraw(&oio_rawterm);
	tcsetattr(0,TCSANOW,&oio_rawterm);
	fputs(CLRSCR,stdout);
}

void oio_finish () {
	tcsetattr(0,TCSANOW,&oio_oldterm);
}

void oio_shoint(int y,int x,const char *prompt,unsigned w,int *a) {
	printf(CSI"%d;%dH%s%*d",y,x,prompt,w,*a);
}

int oio_getint(int y,int x,const char *prompt,unsigned w,int *a) {
	oio_shoint(y,x,prompt,w,a);
	int c = oio_getkey();
	while(c != KEY_ENTER){
		if(c >= '0' && c <= '9') *a = *a*10+(c-'0');
		else if(c==KEY_BKSP) *a /= 10;
		else if(c=='-') *a = -*a;
		else if(c>=' '&&c<077);
		else return c;
		oio_shoint(y,x,prompt,w,a);
		c = oio_getkey();
	}
	printf(TOPLEF CLRCOL);
	return KEY_ENTER;
}


void oio_shostr(int y,int x,const char *prompt,char *str,unsigned w) {
	printf(CSI"%d;%dH%s%-*s"CSI"%dD",y,x,prompt,w,str,w-(int)strlen(str));
}

int oio_getstr(int y,int x,const char *prompt,char *str,unsigned w) {
	int i = strlen(str);
	oio_shostr(y,x,prompt,str,w);
	int c = oio_getkey();
	while(c != KEY_ENTER){
		if(c == KEY_LEFT) return c;
		else if(c == KEY_RIGHT) return c;
		else if(c == KEY_HOME) return c;
		else if(c == KEY_END) return c;
		else if(c == KEY_BKSP && i > 0){
			i--;
			str[i] = 0;
		}
		else if(c == KEY_UP) return c;
		else if(c == KEY_DOWN) return c;
		else if(c >= ' ' && strlen(str) < w){
			str[i] = c;
			i++;
			str[i] = 0;
		}
		oio_shostr(y,x,prompt,str,w);
		c = oio_getkey();
	}
	printf(TOPLEF CLRCOL);
	return KEY_ENTER;
}

int oio_getstredit(int y,int x,const char *prompt,char *str,unsigned w) {
	unsigned i = strlen(str);
	printf(CSI"%d;%dH%s%-*s"CSI"%dD",y,x,prompt,w,str,w-i);
	int c = oio_getkey();
	while(c != KEY_ENTER){
		if(c == KEY_LEFT){
			if(i > 0) i--;
			else return c;
		}else if(c == KEY_RIGHT){
			if(i < strlen(str))i++;
			else return c;
		}else if(c == KEY_HOME) i=0;
		else if(c == KEY_END) i=strlen(str);
		else if(c == KEY_BKSP && i > 0){
			strcpy(str + i-1, str + i);
			i--;
		} else if(c >= ' ' && c < KEY_UP && strlen(str) < w){
			strcpy(str + i+1, str + i);
			str[i] = c;
			i++;
		} else return c;
		printf(CSI"%d;%dH%s%-*s"CSI"%dD",y,x,prompt,w,str,w-i);
		c = oio_getkey();
	}
	printf(TOPLEF CLRCOL);
	return KEY_ENTER;
}

void oio_shohex(int y,int x,const char *prompt,unsigned w,unsigned *a) {
	printf(CSI"%d;%dH%s%*x",y,x,prompt,w,*a);
}

int oio_gethex(int y,int x,const char *prompt,unsigned w,unsigned *a) {
	oio_shohex(y,x,prompt,w,a);
	int c = oio_getkey();
	while(1) {
		if (c >= '0' && c <= '9')*a = *a * 16 + c - '0';
		else if (c >= 'A' && c <= 'F')*a = *a * 16 + c - 'A' + 10;
		else if (c >= 'a' && c <= 'f')*a = *a * 16 + c - 'a' + 10;
		else if(c == KEY_BKSP) *a /= 16;
		else return c;
		oio_shohex(y,x,prompt,w,a);
		c = oio_getkey();
        }
}

int oio_getescseq(){
	int c = oio_getchar();
	if(c == EOF){
		return 033;
	}else if(c == '['){
		int c = oio_getchar();
		if(c=='A')return KEY_UP;
		if(c=='B')return KEY_DOWN;
		if(c=='C')return KEY_RIGHT;
		if(c=='D')return KEY_LEFT;
		if(c=='1') {
			int c = oio_getchar();
			if(c=='~')return KEY_HOME;
			if(c=='1')return KEY_F1;
			oio_ungetc(c);
		}
		if(c=='2') {
			c = oio_getchar();
			if(c=='~')return KEY_INS;
			oio_ungetc(c);
		}
		if(c=='3') {
			int c = oio_getchar();
			if(c=='~')return KEY_DEL;
			oio_ungetc(c);
		}
		if(c=='4') {
			int c = oio_getchar();
			if(c=='~')return KEY_END;
			oio_ungetc(c);
		}
		if(c=='5') {
			int c = oio_getchar();
			if(c=='~')return KEY_PGUP;
			oio_ungetc(c);
		}
		if(c=='6') {
			int c = oio_getchar();
			if(c=='~')return KEY_PGDN;
			oio_ungetc(c);
		}
		oio_ungetc(c);
	}else if(c == 'O'){
		int c = oio_getchar();
		if(c=='A')return KEY_UP;
		if(c=='B')return KEY_DOWN;
		if(c=='C')return KEY_RIGHT;
		if(c=='D')return KEY_LEFT;
		if(c=='F')return KEY_HOME;
		if(c=='H')return KEY_END;
		if(c=='P')return KEY_F1;
		if(c=='Q')return KEY_F2;
		if(c=='R')return KEY_F3;
		if(c=='S')return KEY_F4;
		oio_ungetc(c);
	}
	return c+KEY_ALT;

}


int oio_getkey(){
	int c=u8scanch(oio_getchar,oio_ungetc,0);
	if(c == EOF)return EOF;
	if(c == 033){
		return oio_getescseq();
	}
	return c;
}

void oio_showbar(int total, int cur, int wid) {
	if (total==0)return;
	int i,pix=cur*8*wid/total;
	for(i=0;i<pix/8;i++)fputs("\342\226\210",stdout);
	if(pix%8!=0) printf("\342\226%c",0220-pix%8);
	int bl=wid-pix/8-!!(pix%8);
	for(i=0;i<bl;i++)putchar(' ');
}




