#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "ttf.h"
#include "etc.h"

void write_head_table(FILE *out, int UPM, int xmin, int ymin, int xmax, int ymax){
	write16(out,1);//major
	write16(out,0);//minor
	write32(out,0x00010000);//font version
	write32(out,0);//checksumadjustment
	write32(out,0x5F0F3CF5);//magic number
	write16(out,0x0000);//flags
	write16(out,UPM);//units per em
	//write64(out,time(0) + 24107ULL * 86400ULL);//date created
	//write64(out,time(0) + 24107ULL * 86400ULL);//date modified
	write64(out,0);//for testing purposes, set dates to zero!
	write64(out,0);
	write16(out,xmin);//xmin
	write16(out,ymin);//ymin
	write16(out,xmax);//xmax
	write16(out,ymax);//ymax
	write16(out,0);//macstyle
	write16(out,8);//lowest recommended ppem
	write16(out,2);// deprecated set to 2
	write16(out,1);// 0 short/long 1 format for loca table
	write16(out,0);// glyph data format, always 0
}

void write_hhea_table(FILE *out, int xmax, int ymin, int ymax, unsigned nchars){
	write16(out,1);
	write16(out,0);
	write16(out,ymax);//ascent
	write16(out,ymin);//descent
	write16(out,0);//line gap
	write16(out,xmax);//max advance width
	write16(out,0);//min left sidebearing
	write16(out,0);//min right sidebearing
	write16(out,xmax);//max extent
	write16(out,0);//slope rise
	write16(out,1);//slope run
	write16(out,0);//slant shift
	write16(out,0);//reserved
	write16(out,0);//reserved
	write16(out,0);//reserved
	write16(out,0);//reserved
	write16(out,0);//metric data fmt always 0
	write16(out,nchars);//number of hmtx
}

void write_os_2_table(FILE *out, int UPM, int ymin, int ymax,
		 uint8_t ibm_class, uint8_t ibm_subclass,
		const char *panose, const char *vendorid){
	write16(out,4);
	write16(out,UPM*9/16);//average width
	write16(out,500);//Medium Weight
	write16(out,5);//Medium Width
	write16(out,0);//fstype
	write16(out,UPM);//subscript x size
	write16(out,UPM/2);//subscript y size
	write16(out,0);//subscript x offset
	write16(out,UPM/4);//subscript y offset
	write16(out,UPM);//superscript x size
	write16(out,UPM/2);//superscript y size
	write16(out,0);//superscript x offset
	write16(out,UPM/4);//superscript y offset
	write16(out,UPM*2/16);//strikeout size
	write16(out,UPM/4);//strikeout position
	write8(out,ibm_class);
	write8(out,ibm_subclass);
	/* panose font type */
	int i;
	for (i = 0; i < 10; i++) write8(out,panose[i]);
	/*unicode range bits, set all on for now*/
	write32(out,0xFFFFFFFF);
	write32(out,0xFFFFFFFF);
	write32(out,0xFFFFFFFF);
	write32(out,0xFFFFFFFF);
	/*vendor id*/
	writestr(out,vendorid);// blah
	write16(out,0x1C0);//style bits
	write16(out,0x0020);//minmum character
	write16(out,0xFFFF);//last char
	write16(out,ymax);//ideogram ascension
	write16(out,ymin);//ideogram descent
	write16(out,0);//line gap
	write16(out,ymax);//windows basic ascension
	write16(out,-ymin);//windows basic descent
	/*code page range bits, set all on for now*/
	write32(out,0x7FFFFFFF);
	write32(out,0xFFFFFFFF);
	write16(out,UPM/2);//x height
	write16(out,UPM*3/4);//cap height
	write16(out,0xFFFD);//character to use as replacement
	write16(out,0x0020);//character to use as space
	write16(out,1);//max context of ligatures
}

void write_post_table(FILE * out){
	write32(out,0x00030000);//version 3.0
	write32(out,0x00000000);//upright
	write16(out,0x0000);//underline?
	write16(out,0x0002);//thickness?
	write32(out,0x00000001);//font is monospace
	write32(out,0x00000000);//meomory usage unknown
	write32(out,0x00000000);//meomory usage unknown
	write32(out,0x00000000);//meomory usage unknown
	write32(out,0x00000000);//meomory usage unknown
}

void writenameent(FILE *f,FILE *sf, int plat,int code,int lang, int name,const char *s){
	write16(f,plat);//windows
	write16(f,code);//ucs
	write16(f,lang);//english
	write16(f,name);//name id 1 font family
	int len=strlen(s);
	if(plat==1) {
		write16(f,len);//string length
		write16(f,ftell(sf)); //offset
		fwrite(s,len,1,sf);
	} else if (plat==3) {
		write16(f,len*2);//string length
		write16(f,ftell(sf)); //offset
		writeu16hack(sf,s,len);
	}
}

void write_name_table(FILE * out,const char *name, const char *type, const char *version){
	FILE *strng=fopen("strng.temp","w+");
	write16(out,0);
	write16(out,12);//number of name records
	write16(out,6+12*12);//offset to start of strings
	int scraplen = strlen(name) + 1 + strlen(type) + 1;
	char *scrap = malloc(scraplen);
        strcpy(scrap,name);
        strcat(scrap," ");
        strcat(scrap,type);
	writenameent(out,strng,1,0,0,1,name);
	writenameent(out,strng,1,0,0,2,type);
	writenameent(out,strng,1,0,0,3,scrap);
	writenameent(out,strng,1,0,0,4,name);
	writenameent(out,strng,1,0,0,5,version);
	writenameent(out,strng,1,0,0,6,name);
	writenameent(out,strng,3,1,0x409,1,name);
	writenameent(out,strng,3,1,0x409,2,type);
	writenameent(out,strng,3,1,0x409,3,scrap);
	writenameent(out,strng,3,1,0x409,4,name);
	writenameent(out,strng,3,1,0x409,5,version);
	writenameent(out,strng,3,1,0x409,6,name);
	writefile(out,strng);
}

void write_maxp_table(FILE *out, int nglyph, int maxpts, int maxctr){
	write32(out,0x00010000);
	write16(out,nglyph);//number of glyphs
	write16(out,maxpts);
	write16(out,maxctr);
	write16(out,maxpts);
	write16(out,maxctr);
	write16(out,1);//program maxima
	write16(out,0);
	write16(out,0);
	write16(out,0);
	write16(out,0);
	write16(out,0);
	write16(out,0);
	write16(out,0);
	write16(out,0);
}

void write_cmap_table(FILE *out, FILE *cmap12, unsigned n2groups, unsigned ngroups){
	long cmap_table_start = ftell(out);
	write16(out,0);
	write16(out,2);//2 subtables
	write16(out,3);//platform 3 windows
	write16(out,1);//code 1 unicode BMP
	write32(out,20);//offset
	write16(out,3);//platform 3 windows
	write16(out,10);//code 10 unicode
	long offset_for_cmap12 = ftell(out);
	write32(out,0);//offset
	// offset of 20 is here - begins unicode BMP table
	write16(out,4);//subtable format 4
	unsigned nranges=n2groups + 1;
	write16(out,nranges*8 + 16);//length
	write16(out,0);//language (not used)
	write16(out,nranges*2);
	unsigned searchrange=2;
	unsigned selector=0;
	while(searchrange<=nranges)searchrange*=2,selector++;
	write16(out,searchrange);
	write16(out,selector);
	write16(out,2*nranges - searchrange);
	unsigned i;
	for(i=0;i<n2groups;i++){
		goloc(cmap12,12*i+4);
		write16(out,read32(cmap12));
	}
	write16(out,0xFFFF);
	write16(out,0);
	for(i=0;i<n2groups;i++){
		goloc(cmap12,12*i);
		write16(out,read32(cmap12));
	}
	write16(out,0xFFFF);
	for(i=0;i<n2groups;i++){
		goloc(cmap12,12*i);
		unsigned short x=read32(cmap12);
	        read32(cmap12);
	        x=read32(cmap12) - x;
		write16(out,x);
	}
	write16(out,1);
	for(i=0;i<n2groups;i++){
		write16(out,0);
	}
	write16(out,0);
	unsigned cmap12loc=ftell(out);
	write16(out,12);//subtable format 12
	write16(out,0);//reserved 0
	write32(out,ngroups*12 + 16);//length
	write32(out,0);//language (not used)
	write32(out,ngroups);//number of groups
	writefile(out,cmap12);
	goloc(out,offset_for_cmap12);
	write32(out,cmap12loc-cmap_table_start);
	ffastfw(out);
}


