#include <stdint.h>
#include <stdio.h>
#include "u8.h"
#include <string.h>
#undef getc

/* this software is released to you under the EUPL v. 1.2
   https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12 */

const int win1252[32] = {0x20ac,0x81,0x201A,0x192,0x201E,0x2026,0x2020,0x2021, 0x2C6,0x2030,0x160,0x2039,0x152,0x8D,0x17D,0x8F,
		       0x90,0x2018,0x2019,0x201C,0x201D,0x2022,0x2013,0x2014, 0x2DC,0x2122,0x161,0x203A,0x153,0x9D,0x17E,0x178};

long utf8_my_signature =
(long)("Created by Oren Watson.\n\
this software is released to you under the EUPL v. 1.2\n\
https:/""/joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12\n");

#define isascii(x) ((x)<0200)
#define isinval(x) ((x)>=0370)
#define cont(x) ((x)&077)
#define iscont(x) ((x)>>6==2)
#define isncont(x) ((x)>>6!=2)
#define isstart(x) ((x)>=0300&&(x)<0370)
#define isnstart(x) ((x)<0300||(x)>=0370)
#define is2start(x) ((x)>>5==06)
#define start2(x) ((x)&037)
#define is3start(x) ((x)>>4==016)
#define start3(x) ((x)&017)
#define is4start(x) ((x)>>3==036)
#define start4(x) ((x)&07)
#define tounich2(s,i) (start2(s[i])<<6|cont(s[i+1]))
#define tounich3(s,i) (start3(s[i])<<12|cont(s[i+1])<<6|cont(s[i+2]))
#define tounich4(s,i) (start4(s[i])<<18|cont(s[i+1])<<12\
			|cont(s[i+2])<<6|cont(s[i+3]))
#define tounich5(s,i) ((s[i]&03)<<24|cont(s[i+1])<<18\
			|cont(s[i+2])<<12|cont(s[i+3])<<6|cont(s[i+4]))

int u8chlen(unsigned ch){
	if(ch<0200)return 1;
	if(ch<04000)return 2;
	if(ch<0200000)return 3;
	return 4;
}

char *u8chtostr(unsigned ch){
	static char buf[5];
	buf[4]=0;
	if(ch<0200){
		buf[3]=ch;
		return buf+3;
	}
	if(ch<04000){
		buf[3]=(ch&077) + 0200;
		buf[2]=(ch/0100) + 0300;
		return buf+2;
	}
	if(ch<0200000){
		buf[3]=(ch&077) + 0200;
		buf[2]=(ch/0100&077) + 0200;
		buf[1]=(ch/010000) + 0340;
		return buf+1;
	}
	buf[3]=(ch&077) + 0200;
	buf[2]=(ch/0100&077) + 0200;
	buf[1]=(ch/010000&077) + 0200;
	buf[0]=(ch/01000000) + 0360;
	return buf;
}

int u8fputch(FILE *f, unsigned ch){
	return fputs(u8chtostr(ch),f);
}

int u8putch(unsigned ch){
	return fputs(u8chtostr(ch),stdout);
}

int u8findprev(const char *s_, int i){
	const unsigned char *s = (const unsigned char*)s_;
	//todo: handle invalid utf8 at all
	if(i <= 0)return -1;
	int j = i-1;
	while(1){
		if(iscont(s[j]) && j > 0) j--;
		else return j;
	}
}

int u8findnextn(const char *s_, int i, int n){
	const unsigned char *s = (const unsigned char*)s_;
	if(i>=n)return -1;
	int k;
	if(s[i]<0300)return i+1;
	if(s[i]<0340)k=i+2;
	else if(s[i]<0360)k=i+3;
	else if(s[i]<0370)k=i+4;
	else return i+1;
	int j=i+1;
	while(j<k){
		if((s[j]&0300)!=0200)return i+1;
		j++;
	}
	return j;
}

int u8findnext(const char *s, int i){
	return u8findnextn(s,i,strlen(s));
}

int u8scancont(int (*getc)(void *f),void (*ungetc)(char,void*),void*f,int c,int n){
	int cs[4];//char stack
	int i=0;
	do{cs[i]=getc(f);
	if(cs[i]==EOF)goto fail;
	i++;
	if(cs[i-1]<0200||cs[i-1]>=0300)goto fail;
	}while(i<n);
	for(i=0;i<n;i++)c=c*0100+(cs[i]&077);
	if(0)fail:while(i>0)ungetc(cs[--i],f);
	return c;
}

int u8_scanch(int(*getc)(void *f),void(*ungetc)(char,void*),void*f){
	int c=getc(f);
	if(c==EOF)return EOF;
	if(c<0200)return c;
	if(c<0240)return win1252[c-128];// invalid start byte. fallback to windows1252
	if(c<0300)return c; // invalid start byte. fallback to latin-1
	if(c<0340)return u8scancont(getc,ungetc,f,c&077,1);
	if(c<0360)return u8scancont(getc,ungetc,f,c-0340,2);
	if(c<0370)return u8scancont(getc,ungetc,f,c-0360,3);
	return c;//start bytes for supernal plane, fall back to latin 1
}

int u8fscanch(FILE *f){
	return u8_scanch((int(*)(void*))fgetc,(void(*)(char,void*))ungetc,f);
}

int u8scanch(void){
	return u8_scanch((int(*)(void*))fgetc,(void(*)(char,void*))ungetc,stdin);
}

static int cstrppgetc(const char **s){
	if(**s==0)return EOF;
	int c=**(const unsigned char**)s;
	(*s)++;
	return c;
}

static void cstrppungetc(char c, const char **s){
	s[0]--;
}

int u8sscanch(const char *s,const char **end){
	int c = u8_scanch((int(*)(void*))cstrppgetc,(void(*)(char,void*))cstrppungetc,&s);
	if(end!=0)*end=s;
	return c;
}

struct memblockwithindex {
	const unsigned char *s;
	int i;
	int n;
};

static int memblockgetc(struct memblockwithindex*s){
	if(s->i>=s->n)return EOF;
	return s->s[s->i++];
}

static void memblockungetc(char c, struct memblockwithindex*s){
	s->i--;
}

int u8memscanch(const char *s,int i,int len,int *end){
	struct memblockwithindex mem;
	mem.s=s;
	mem.i=i;
	mem.n=len;
	int c = u8_scanch((int(*)(void*))memblockgetc,(void(*)(char,void*))memblockungetc,&mem);
	if(end!=0)*end=mem.i;
	return c;
}

// approximation of width of 32 bit character
int u8apxwid(unsigned ch){
	if (ch < 0x1100)return 1;
	if (ch < 0x1200)return 2;
	if (ch < 0x2460)return 1;
	if (ch < 0x2500)return 2;
	if (ch < 0x2E80)return 1;
	if (ch < 0xA4D0)return 2;
	return 1;
}

static inline int findzero4(const char *s, int i){
	int z=i;
	while(z-i<4&&s[z]!=0)z++;
	return z;
}

int u8prevch(const char *s_, int i){
	const unsigned char *s = (const unsigned char*)s_;
	if(i<=0)return -1;
	if(isncont(s[i-1])||i-1==0)return i-1;
	if(is2start(s[i-2]))return i-2;//valid 2 byte char
	if(isncont(s[i-2])||i-2==0)return i-1;// need another byte.
	if(is3start(s[i-3]))return i-3;//valid 3 byte char
	if(isncont(s[i-3])||i-3==0)return i-1;
	if(is4start(s[i-4]))return i-4;//valid 4 byte char
	return i-1;
}

int u8resync(const char *s_, int i){
	const unsigned char *s = (const unsigned char*)s_;
	if(i<=0)return 0;
	if(!iscont(s[i]))return i;
	if(is2start(s[i-1]))return i-1;
	if(i-1==0)return i;
	if(!iscont(s[i-1]))return i;
	if(is3start(s[i-2]))return i-2;
	if(i-2==0)return i;
	if(!iscont(s[i-2]))return i;
	if(is4start(s[i-3]))return i-3;
	return i;
}

int u8nextch(const char *s_, int i, int n){
	const unsigned char *s = (const unsigned char*)s_;
	if(i>=n)return -1;
	if(isnstart(s[i])||i+1==n||isncont(s[i+1]))return i+1;
	if(is2start(s[i]))return i+2;
	if(i+2==n||isncont(s[i+2]))return i+1;
	if(is3start(s[i]))return i+3;
	if(i+3==n||isncont(s[i+3]))return i+1;
	return i+4;
}

int u8snextch(const char *s, int i){
	return u8nextch(s,i,findzero4(s,i));
}

int u8ithchr(const char *_s, int ci, int n){
	const unsigned char *s = _s;
	int i = 0;
	int c = 0;
	while(i<n&&i!=-1&&c!=ci)i=u8nextch(s,i,n),c++;
	return i;
}

int u8sithchr(const char *s, int ci){
	return u8ithchr(s,ci,strlen(s));
}

int u8chrcnt(const char *_s, int z){
	const unsigned char *s = _s;
	int i = 0;
	int c = 0;
	while(i<z){
		if(isnstart(s[i])){c++;i++;}
		else if(i+1==z)return c+1;
		else if(isncont(s[i+1])){c+=2;i+=2;}
		else if(is2start(s[i])){c++;i+=2;}
		else if(i+2==z)return c+2;
		else if(isncont(s[i+2])){c+=3;i+=3;}
		else if(is3start(s[i])){c++;i+=3;}
		else if(i+3==z)return c+3;
		else if(isncont(s[i+3])){c+=4;i+=4;}
		else{c++;i+=4;}
	}
	return c;

}

int u8schrcnt(const char *s){
	return u8chrcnt(s,strlen(s));
}

unsigned u8rdnxchr(const char *s_, int i, int n){
	const unsigned char *s = (const unsigned char*)s_;
	if(i>=n)return UTF8_EOSTR;
	if(isnstart(s[i])||i+1==n||isncont(s[i+1]))return s[i];
	if(is2start(s[i]))return tounich2(s,i);
	if(i+2==n||isncont(s[i+2]))return s[i];
	if(is3start(s[i]))return tounich3(s,i);
	if(i+3==n||isncont(s[i+3]))return s[i];
	return tounich4(s,i);
}

unsigned u8srdnxchr(const char *s, int i){
	return u8rdnxchr(s,i,findzero4(s,i));
}

unsigned u8ith(const char *s, int i, int n){
	i = u8ithchr(s,i,n);
	if(i < 0)return UTF8_EOSTR;
	return u8rdnxchr(s,i,n);
}

unsigned u8sith(const char *s, int i){
	int n = strlen(s);
	i = u8ithchr(s,i,n);
	if(i < 0)return UTF8_EOSTR;
	return u8rdnxchr(s,i,n);
}

unsigned u8strtoch(const char *s){
	return u8rdnxchr(s,0,findzero4(s,0));
}

unsigned u8classify(const char *s_, int i, int n){
	const unsigned char *s = (const unsigned char*)s_;
	if(i>=n)return UTF8CLASS_EOSTR;
	if(isascii(s[i]))return UTF8CLASS_ASCII|UTF8CLASS_VALID;
	if(iscont(s[i]))return UTF8CLASS_CONTI|UTF8CLASS_INVAL;
	if(is2start(s[i])){
		if(i+1==n||isncont(s[i+1]))
			return UTF8CLASS_START|UTF8CLASS_INVAL;
		return UTF8CLASS_2BYTE|UTF8CLASS_VALID;
	}
	if(is3start(s[i])){
		if(i+2>=n||isncont(s[i+1])||isncont(s[i+2]))
			return UTF8CLASS_START|UTF8CLASS_INVAL;
		return UTF8CLASS_3BYTE|UTF8CLASS_VALID;
	}
	if(is4start(s[i])){
		if(i+3==n||isncont(s[i+1])||isncont(s[i+2])||isncont(s[i+3]))
			return UTF8CLASS_START|UTF8CLASS_INVAL;
		return UTF8CLASS_4BYTE|UTF8CLASS_VALID;
	}
	return UTF8CLASS_INVAL;
}

unsigned u8sclassify(const char *s, int i){
	return u8classify(s,i,findzero4(s,i));
}
