#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "unidat.h"

int unidat_fd;
void *unidat;
const char **uni_names;

struct eaw_entry {
	int lb; // lower bound
	int ub; // upper bound
	int w;  // effective width
};

unsigned eawdat_n = 0;
struct eaw_entry *eawdat = NULL;

int is_HEX(char x){
	return (x >= '0' && x <= '9') || (x >= 'A' && x <= 'F');
}

int dig_val(char x){
	if(x >= '0' && x <= '9')return x-'0';
	if(x >= 'A' && x <= 'Z')return x-'A'+10;
	if(x >= 'a' && x <= 'z')return x-'a'+10;
	return 0;
}

int load_eawdat() {
	FILE *eawdat_f = fopen("EastAsianWidth.txt","r");
	if (eawdat_f == NULL) return 1;
	char line[200];
	eawdat = malloc(sizeof(struct eaw_entry)*3000);
	int i = 0;
	int lb,ub,w;
	char wc;
	int last_lb = -1;
	int last_w = -1;
	while (fgets(line,200,eawdat_f)) {
		if(i == 3000){
			printf("ran out of eaw space\n");
			abort();
		}
		if(line[0] == '#' || line[0] == '\n')continue;
		int j = 4;
		while(is_HEX(line[j]))j++;
		if(line[j] == ';'){
			if (2 != sscanf(line,"%x;%c",&lb,&wc)){
				printf("failed! %s\n",line);
                                abort();
			};
			ub = lb;
		}
		if(line[j] == '.'){
			if (3 != sscanf(line,"%x..%x;%c",&lb,&ub,&wc)){
				printf("failed! %s\n",line);
				abort();
			}
		}
		w = 1 + (wc == 'F' || wc == 'W');
		while(line[j] != '#' && line[j] != '\n')j++;
		if(line[j] == '\n'){
			printf("failed! %s\n",line);
			abort();
		}
		char cc1, cc2;
		if (2 != sscanf(line+j,"# %c%c",&cc1,&cc2)){
			printf("failed! %s\n",line);
			abort();
		}
		if(cc1=='M'&&(cc2=='n'||cc2=='e'))w=0;
		eawdat[i].lb = lb;
		eawdat[i].ub = ub;
		eawdat[i].w = w;
		//printf("%d %x %x %d\n",i,eawdat[i].lb, eawdat[i].ub, eawdat[i].w);
		i++;
	}
	eawdat_n = i;
	fclose(eawdat_f);
	return i == 0;
}

int get_eaw(unsigned ch_u) {
	if(eawdat_n == 0 || eawdat == NULL){
		printf("eaw not loaded!\n");
		abort();
	}
	int ch = ch_u;
	int i = eawdat_n/2;
	int t = eawdat_n;
	int b = -1;
	while (1) {
		if (i == t || i == b)return 1;
		if (ch > eawdat[i].ub) {
			b = i;
			i = (i + t)/2;
		} else if (ch < eawdat[i].lb) {
			t = i;
			i = (i + b)/2;
		} else {
			return eawdat[i].w;
		}
	}
}

int unload_eawdat() {
	free(eawdat);
	eawdat_n = 0;
	return 0;
}

int write_eawbin() {
	FILE *f = fopen("eaw.bin","w");
	if(!f)return 1;
	fprintf(f,"east asian width binary data\nentries:%d\n",eawdat_n);
	fwrite(eawdat, sizeof(struct eaw_entry), eawdat_n, f);
	fclose(f);
	return 0;
}

int load_eawbin() {
	FILE *f = fopen("/usr/share/eaw.bin","r");
	if(!f)f = fopen("/usr/local/share/eaw.bin","r");
	if(!f)f = fopen("eaw.bin","r");
	if(!f)return 1;
	char line[200];
	if(fgets(line,200,f) == NULL)goto fail;
	if(fgets(line,200,f) == NULL)goto fail;
	if(1 != sscanf(line,"entries:%d",&eawdat_n))goto fail;
	eawdat = malloc(sizeof(struct eaw_entry)*eawdat_n);
	if(fread(eawdat, sizeof(struct eaw_entry), eawdat_n, f) != eawdat_n)goto fail;
	fclose(f);
	return 0;
	fail:
	if(eawdat)free(eawdat);
	if(f)fclose(f);
	return 1;
}

int load_unidat() {
	unidat_fd = open("UnicodeData.txt",O_RDONLY);
	if(unidat_fd == -1)return 1;
	off_t end = lseek(unidat_fd, 0, SEEK_END);
	unidat = mmap(NULL, end, PROT_READ, MAP_SHARED, unidat_fd, 0);
	if(unidat == MAP_FAILED){
		close(unidat_fd);
		return 1;
	}
	uni_names = (void*)calloc(0x20000,sizeof(char*));
	const char *base = unidat;
	const char *s = base;
	while(1){
		int c = 0;
		while(is_HEX(*s)) {
			c = c*16 + dig_val(*s);
			s++;
		}
		if(c >= 0x20000)break;
		if(s >= base + end)break;
		s++;
		uni_names[c] = s;
		while(*s != '\n')s++;
		s++;
	}
	return 0;
}

const char *get_unidat_entry(unsigned ch) {
	if (ch >= 0x20000) return "over_name_limit\n";
	return uni_names[ch];
}

int unload_unidat() {
        free(uni_names);
	off_t end = lseek(unidat_fd, 0, SEEK_END);
	munmap(unidat,end);
        close(unidat_fd);
	return 0;
}


